#include <sys/keycodes.h>

#include <assert.h>
#include <display.h>
#include <err.h>
#include <errno.h>
#include <pixel.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <timespec.h>

#define STAGES 8
enum game_mode { TITLESCREEN, STAGE };
enum game_mode game_mode = TITLESCREEN;
bool stages_beat[STAGES];
size_t stages_num;

#define PLAYFIELD_SIDE 100
#define TILE_SIDE 10
#define TILES (PLAYFIELD_SIDE / TILE_SIDE)
enum palette {
	BG = 0,
	TEXT,
	PLAYER_OUTLINE,
	PLAYER_SKIN,
	PLAYER_EYE,
	GRASS_TOP,
	GRASS_BROWN,
	GRASS_DARK,
	FLAG_STEM,
	FLAG_FLAG,
	SWITCH_SWITCH,
	TILE1,
	TILE2,
	TILE3,
	TILE4,
	STAR_STAR,
};
static enum palette playfield[PLAYFIELD_SIDE * PLAYFIELD_SIDE];

#include "player_stand_tile.inc"
#include "player_walk1_tile.inc"
#include "player_walk2_tile.inc"
#include "player_walk1_tile_reversed.inc"
#include "player_walk2_tile_reversed.inc"

#include "grass_tile.inc"
#include "dirt_tile.inc"
#include "flag_tile.inc"

#include "switch1_off_tile.inc"
#include "switch1_on_tile.inc"
#include "switch2_off_tile.inc"
#include "switch2_on_tile.inc"
#include "switch3_off_tile.inc"
#include "switch3_on_tile.inc"
#include "switch4_off_tile.inc"
#include "switch4_on_tile.inc"
#include "switch0_off_tile.inc"
#include "switch0_on_tile.inc"

#include "tile1_off_tile.inc"
#include "tile1_on_tile.inc"
#include "tile2_off_tile.inc"
#include "tile2_on_tile.inc"
#include "tile3_off_tile.inc"
#include "tile3_on_tile.inc"
#include "tile4_off_tile.inc"
#include "tile4_on_tile.inc"

#include "stage1_tile.inc"
#include "stage2_tile.inc"
#include "stage3_tile.inc"
#include "stage4_tile.inc"
#include "stage5_tile.inc"
#include "stage6_tile.inc"
#include "stage7_tile.inc"
#include "stage8_tile.inc"
#include "star_tile.inc"

#include "rebind_rebind1_tile.inc"
#include "rebind_rebind2_tile.inc"
#include "rebind_rebind3_tile.inc"
#include "rebind_left1_tile.inc"
#include "rebind_left2_tile.inc"
#include "rebind_right1_tile.inc"
#include "rebind_right2_tile.inc"
#include "rebind_jump1_tile.inc"
#include "rebind_jump2_tile.inc"
#include "rebind_rebind1_alt_tile.inc"

static char tilemap[TILES * TILES + 1];
static char *stages[STAGES];
static char *builtin_stages[] = {
	"          "
	"          "
	"          "
	"!  g      "
	"gg    g   "
	"         g"
	"        g "
	"     g    "
	"@   gdg   "
	"gg  dddgg ",
	"          "
	"          "
	"          "
	"          "
	"          "
	"          "
	" A        "
	"    11g   "
	"@        !"
	"ggg     gg",
	"          "
	"          "
	"B        N"
	"2222222222"
	"  1       "
	"          "
	"1         "
	"          "
	"@A1      !"
	"g1      gg",
	" d        "
	" d    N   "
	" d        "
	" d  g     "
	" d        "
	"Adg    ggg"
	"      1  !"
	"     g   g"
	"    gd  gd"
	"ggggddggdd",
	"         !"
	"     D   4"
	"C   3   44"
	"3  3 4    "
	"         B"
	" 222  2  2"
	"          "
	" A  1  111"
	"@g        "
	"gd        ",
	"     @    "
	"AggggggggB"
	"     1    "
	"     1    "
	"1ggBgggC11"
	"          "
	"gg2ggg3333"
	"          "
	"   !      "
	"  ggg     ",
	"          "
	"   C     N"
	"          "
	"   B    3 "
	"   2 3    "
	"1  3     1"
	"        12"
	"@ 1    123"
	"A1    123!"
	"1     123g",
	"@d   B12 C"
	"A     1ggg"
	"11gggggd  "
	"         D"
	"gg44444444"
	"      22  "
	"     2    "
	"  2      2"
	"        2!"
	"    333333",
};

enum switches_state { ALL_OFF, SWITCH1, SWITCH2, SWITCH3, SWITCH4 };
enum switches_state switches_state;

static uint32_t main_window = 0;
static uint32_t window_width = 500;
static uint32_t window_height = 500;

static bool game_running = true;

static void on_disconnect(void *context) {
	(void) context;
	game_running = false;
}

static bool left_held, right_held, jump_held;

static double jump_maxheight = 21;
static double jump_duration = 0.4;

static int left_key = KBKEY_LEFT;
static int right_key = KBKEY_RIGHT;
static int jump_key = KBKEY_SPACE;
static int rebind_key = KBKEY_R;

enum rebind_state { PLAYING, LEFT, RIGHT, JUMP, REBIND };
enum rebind_state rebind_state;

static void on_keyboard(void *context, uint32_t window, uint32_t codepoint) {
	(void) context;
	if (window != main_window)
		return;
	int key = KBKEY_DECODE(codepoint);
	switch (rebind_state) {
		case PLAYING:
			if (key == -rebind_key) rebind_state = LEFT;
			else if (key == left_key) left_held = true;
			else if (key == -left_key) left_held = false;
			else if (key == right_key) right_held = true;
			else if (key == -right_key) right_held = false;
			else if (key == jump_key) jump_held = true;
			else if (key == -jump_key) jump_held = false;
			break;
		case LEFT:
			if (key >= 0) return;
			left_key = -key;
			rebind_state = RIGHT;
			break;
		case RIGHT:
			if (key >= 0) return;
			right_key = -key;
			rebind_state = JUMP;
			break;
		case JUMP:
			if (key >= 0) return;
			jump_key = -key;
			rebind_state = REBIND;
			break;
		case REBIND:
			if (key >= 0) return;
			rebind_key = -key;
			rebind_state = PLAYING;
			break;
	}
}

static void on_quit(void *context, uint32_t window) {
	(void) context;
	if (window != main_window)
		return;
	game_running = false;
}

static void on_resize(void *context, uint32_t window, uint32_t width, uint32_t height) {
	(void) context;
	if (window != main_window)
		return;
	window_width = width;
	window_height = height;
}

static double timespec2double(struct timespec ts) {
	return ts.tv_sec + ts.tv_nsec / 1000.0 / 1000.0 / 1000.0;
}

static char sample_tilemap(size_t x, size_t y) {
	size_t tx = x / TILE_SIDE;
	size_t ty = y / TILE_SIDE;
	if (tx >= TILES || ty >= TILES) return ' ';
	return tilemap[ty * TILES + tx];
}

static bool solid_tile(size_t x, size_t y) {
	char tile = sample_tilemap(x, y);
	return tile != ' ' && tile != '!' && tile != '@' &&
           tile != 'A' && (tile != '1' || switches_state == SWITCH1) &&
           tile != 'B' && (tile != '2' || switches_state == SWITCH2) &&
           tile != 'C' && (tile != '3' || switches_state == SWITCH3) &&
           tile != 'D' && (tile != '4' || switches_state == SWITCH4) &&
	       tile != 'N';
}

static bool jumping = false;
static bool on_ground = false;
static double jump_start_y;
static double in_jump;
static double player_x, player_y;
static double player_dx, player_dy;

static size_t selection_index = 0;

static void initialize_stage(void) {
	strcpy(tilemap, stages[selection_index]);
	switches_state = ALL_OFF;
	jumping = false;
	on_ground = false;
	player_dx = 0;
	player_x = 0;
	player_y = 0;
	for (size_t ty = 0; ty < TILES; ty++) {
		for (size_t tx = 0; tx < TILES; tx++) {
			if (tilemap[ty * TILES + tx] == '@') {
				player_x = tx * TILE_SIDE;
				player_y = ty * TILE_SIDE;
			}
		}
	}
}

static bool keys_released;
static bool stage_selected; 

static void initialize_titlescreen(void) {
	keys_released = false; stage_selected = false;
}

static void initialize(void) {
	switch (game_mode) {
		case TITLESCREEN: initialize_titlescreen(); break;
		case STAGE: initialize_stage(); break;
		default: printf("game_mode %i\n", game_mode);
	}
}

static void update_stage(struct timespec now, struct timespec dt_timespec) {
	(void) now;
	double dt = timespec2double(dt_timespec);
	player_x += player_dx * dt;

	if (left_held) player_dx = -30;
	else if (right_held) player_dx = 30;
	else player_dx = 0;

	if (on_ground && jump_held && !jumping) {
		jumping = true;
		in_jump = 0;
		jump_start_y = player_y;
	}
	if (!on_ground && !jumping) {
		jumping = true;
		in_jump = jump_duration;
		jump_start_y = player_y + jump_maxheight;
	}
	if (jumping) {
		in_jump += dt;
		// Jump is a quadratic arc
		// h = at^2 + bt + c
		//
		// Beginning of jump, zero height
		// t = 0, h = 0
		// 0 = a0^2 + b0 + c
		// c = 0
		//
		// At 2*duration, zero height
		// t = 2T, h = 0
		// 0 = a(2T)^2 + b(2T) + 0
		// 0 = 4aT^2 + 2bT
		// 0 = 2aT^2 + bT
		// 0 = 2aT + b
		// b = -2aT
		//
		// At duration, maximum height
		// t = T, h = H
		// H = aT^2 + bT + 0
		// H = aT^2 - 2aT^2
		// H = -aT^2
		// a = -H / (T^2)
		// b
		double a = -jump_maxheight / (jump_duration * jump_duration);
		double b = -2 * a * jump_duration;
		double jump_height = a * in_jump * in_jump + b * in_jump;
		player_y = jump_start_y - jump_height;
		// player_dy is negative of derivative of jump_height
		// H = a*t^2 + b*t
		// H' = 2*a*t + b
		player_dy = -2 * a * in_jump - b;
	}

	// Collision against walls of the map
	if (player_x < 0) {
		player_x = 0;
		player_dx = 0;
	} else if (player_x > PLAYFIELD_SIDE - TILE_SIDE) {
		player_x = PLAYFIELD_SIDE - TILE_SIDE;
		player_dx = 0;
	}
	if (player_y < 0) {
		player_y = 0;
		player_dy = 0;
	} else if (player_y > PLAYFIELD_SIDE - TILE_SIDE) {
		player_y = PLAYFIELD_SIDE - TILE_SIDE;
		player_dy = 0;
		game_mode = TITLESCREEN;
		initialize();
		return;
	}

	// Collision against tiles
	size_t px = player_x;
	size_t py = player_y;
	size_t topbottom_left_cx = px + 1;
	size_t topbottom_right_cx = px + TILE_SIDE - 1 - 1;
	size_t top_cy = py - 1;
	size_t bottom_cy = py + TILE_SIDE;

	on_ground = false;
	if (solid_tile(topbottom_left_cx, bottom_cy) ||
	    solid_tile(topbottom_right_cx, bottom_cy)) {
		on_ground = true;
		if (player_dy > 0) {
			jumping = false;
			size_t snapped_py = py / TILE_SIDE * TILE_SIDE;
			player_y = snapped_py;
		}
	}
	if ((solid_tile(topbottom_left_cx, top_cy) ||
	     solid_tile(topbottom_right_cx, top_cy)) &&
		player_dy < 0) {
		size_t snapped_py = (py + TILE_SIDE - 1) / TILE_SIDE * TILE_SIDE;
		player_y = snapped_py;
		in_jump = jump_duration;
		jump_start_y = player_y + jump_maxheight;
	}

	px = player_x;
	py = player_y;
	size_t left_cx = px - 1;
	size_t right_cx = px + TILE_SIDE;
	size_t leftright_top_cy = py;
	size_t leftright_bottom_cy = py + TILE_SIDE - 3; //XXX: proper handling of snapping
	if ((solid_tile(left_cx, leftright_top_cy) ||
	     solid_tile(left_cx, leftright_bottom_cy)) &&
		player_dx < 0) {
		size_t snapped_px = (px + TILE_SIDE - 1) / TILE_SIDE * TILE_SIDE;
		player_x = snapped_px;
		player_dx = 0;
	}
	if ((solid_tile(right_cx, leftright_top_cy) ||
	     solid_tile(right_cx, leftright_bottom_cy)) &&
	    player_dx > 0) {
		size_t snapped_px = px / TILE_SIDE * TILE_SIDE;
		player_x = snapped_px;
		player_dx = 0;
	}

	// Touching tiles
	size_t midpoint_x = player_x + TILE_SIDE / 2;
	size_t midpoint_y = player_y + TILE_SIDE / 2;
	switch(sample_tilemap(midpoint_x, midpoint_y)) {
		case '!':
			stages_beat[selection_index] = true;
			if (selection_index + 1 < stages_num)
				selection_index++;
			game_mode = TITLESCREEN;
			initialize();
			return;
		case 'A':
			switches_state = SWITCH1;
			break;
		case 'B':
			switches_state = SWITCH2;
			break;
		case 'C':
			switches_state = SWITCH3;
			break;
		case 'D':
			switches_state = SWITCH4;
			break;
		case 'N':
			switches_state = ALL_OFF;
	}
}

static void update_titlescreen(struct timespec now, struct timespec dt_timespec) {
	(void) now;
	(void) dt_timespec;
	if (keys_released && left_held && selection_index > 0)
		selection_index--;
	if (keys_released && right_held && selection_index + 1 < stages_num)
		selection_index++;
	if (keys_released && jump_held)
		stage_selected = true;
	if (!left_held && !right_held && !jump_held)
		keys_released = true;
	else
		keys_released = false;
	if (stage_selected && keys_released) {
		game_mode = STAGE;
		initialize();
	}
}

static void update(struct timespec now, struct timespec dt_timespec) {
	if (rebind_state != PLAYING) return;
	switch (game_mode) {
		case TITLESCREEN: update_titlescreen(now, dt_timespec); break;
		case STAGE: update_stage(now, dt_timespec); break;
		default: printf("game_mode %i\n", game_mode);
	}
}

static void draw_tile(enum palette tile[], size_t x, size_t y) {
	assert(x + TILE_SIDE <= PLAYFIELD_SIDE);
	assert(y + TILE_SIDE <= PLAYFIELD_SIDE);
	for (size_t ty = 0; ty < TILE_SIDE; ty++) {
		for (size_t tx = 0; tx < TILE_SIDE; tx++) {
			enum palette pixel = tile[ty * TILE_SIDE + tx];
			if (pixel != BG)
				playfield[(y + ty) * PLAYFIELD_SIDE + x + tx] = pixel;
		}
	}
}

static void draw_tiles(void) {
	assert(strlen(tilemap) == TILES * TILES);
	for (size_t ty = 0; ty < TILES; ty++) {
		for (size_t tx = 0; tx < TILES; tx++) {
			enum palette *tile = NULL;
			switch(tilemap[ty * TILES + tx]) {
				case 'g': tile = grass_tile; break;
				case 'd': tile = dirt_tile; break;
				case '!': tile = flag_tile; break;
				case 'A': tile = switches_state == SWITCH1 ? switch1_on_tile : switch1_off_tile; break;
				case 'B': tile = switches_state == SWITCH2 ? switch2_on_tile : switch2_off_tile; break;
				case 'C': tile = switches_state == SWITCH3 ? switch3_on_tile : switch3_off_tile; break;
				case 'D': tile = switches_state == SWITCH4 ? switch4_on_tile : switch4_off_tile; break;
				case 'N': tile = switches_state == ALL_OFF ? switch0_on_tile : switch0_off_tile; break;
				case '1': tile = switches_state == SWITCH1 ? tile1_on_tile : tile1_off_tile; break;
				case '2': tile = switches_state == SWITCH2 ? tile2_on_tile : tile2_off_tile; break;
				case '3': tile = switches_state == SWITCH3 ? tile3_on_tile : tile3_off_tile; break;
				case '4': tile = switches_state == SWITCH4 ? tile4_on_tile : tile4_off_tile; break;
			}
			if (tile)
				draw_tile(tile, tx * TILE_SIDE, ty * TILE_SIDE);
		}
	}
}

static void draw_stage(struct timespec now) {
	draw_tiles();
	if (right_held && now.tv_nsec % 500000000L < 250000000L && on_ground)
		draw_tile(player_walk2_tile, player_x, player_y);
	else if (right_held)
		draw_tile(player_walk1_tile, player_x, player_y);
	else if (left_held && now.tv_nsec % 500000000L < 250000000L && on_ground)
		draw_tile(player_walk2_tile_reversed, player_x, player_y);
	else if (left_held)
		draw_tile(player_walk1_tile_reversed, player_x, player_y);
	else
		draw_tile(player_stand_tile, player_x, player_y);
}

static void draw_titlescreen(struct timespec now) {
	(void) now;
	for (size_t i = 0; i < stages_num; i++) {
		size_t x = i * TILE_SIDE + TILE_SIDE;
		size_t y = 0;
		if (stages_beat[i])
			draw_tile(star_tile, x, y);
	}
	if (stages_num >= 1) draw_tile(stage1_tile, 1 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 2) draw_tile(stage2_tile, 2 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 3) draw_tile(stage3_tile, 3 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 4) draw_tile(stage4_tile, 4 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 5) draw_tile(stage5_tile, 5 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 6) draw_tile(stage6_tile, 6 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 7) draw_tile(stage7_tile, 7 * TILE_SIDE, TILE_SIDE);
	if (stages_num >= 8) draw_tile(stage8_tile, 8 * TILE_SIDE, TILE_SIDE);
	size_t underline_y = 2 * TILE_SIDE + 1;
	size_t underline_x = selection_index * TILE_SIDE + TILE_SIDE + 2;
	for (size_t x = 0; x < TILE_SIDE - 4; x++)
		playfield[underline_y * PLAYFIELD_SIDE + underline_x + x] = TEXT;
}

static void draw(struct timespec now) {
	memset(playfield, 0, sizeof(playfield));
	if (rebind_state == PLAYING) {
		switch (game_mode) {
			case TITLESCREEN: draw_titlescreen(now); break;
			case STAGE: draw_stage(now); break;
			default: printf("game_mode %i\n", game_mode);
		}
	} else {
		draw_tile(rebind_rebind1_tile, 0 * TILE_SIDE, 0);
		draw_tile(rebind_rebind2_tile, 1 * TILE_SIDE, 0);
		draw_tile(rebind_rebind3_tile, 2 * TILE_SIDE, 0);
		switch (rebind_state) {
			case LEFT:
				draw_tile(rebind_left1_tile, 3 * TILE_SIDE, 0);
				draw_tile(rebind_left2_tile, 4 * TILE_SIDE, 0);
				break;
			case RIGHT:
				draw_tile(rebind_right1_tile, 3 * TILE_SIDE, 0);
				draw_tile(rebind_right2_tile, 4 * TILE_SIDE, 0);
				break;
			case JUMP:
				draw_tile(rebind_jump1_tile, 3 * TILE_SIDE, 0);
				draw_tile(rebind_jump2_tile, 4 * TILE_SIDE, 0);
				break;
			case REBIND:
				draw_tile(rebind_rebind1_alt_tile, 3 * TILE_SIDE, 0);
				draw_tile(rebind_rebind2_tile, 4 * TILE_SIDE, 0);
				draw_tile(rebind_rebind3_tile, 5 * TILE_SIDE, 0);
				break;
			default:
				printf("rebind_state %i\n", rebind_state);
		}
	}
}

static void load_stages(char *path) {
	FILE *f = fopen(path, "r");
	if (!f)
		err(1, "%s", path);

	stages_num = 0;

	bool in_comment = true;
	size_t linenum = 0;
	size_t stage_fill = 0;
	char *stage[10];
	while (true) {
		linenum++;
		char *line = NULL;
		size_t size = 0;
		ssize_t length = getline(&line, &size, f);
		if (length < 0) {
			break;
			free(line);
		}
		if (line[length - 1] == '\n') {
			line[length - 1] = 0;
			length--;
		}
		if (!strcmp(line, "----------")) {
			if (in_comment) {
				in_comment = false;
				continue;
			} else {
				if (stages_num == STAGES)
					errx(1, "%s: %zu: Too many stages (maximum %zu)", path, linenum, (size_t)STAGES);
				free(line);
				if (stage_fill < 10)
					errx(1, "%s: %zu: Not enough lines for a stage (expected 10, got %zu)", path, linenum, stage_fill);	
				char *stage_str;
				asprintf(&stage_str, "%s%s%s%s%s%s%s%s%s%s", stage[0], stage[1], stage[2], stage[3], stage[4], stage[5], stage[6], stage[7], stage[8], stage[9]);
				stages[stages_num++] = stage_str;
				for (size_t i = 0; i < stage_fill; i++)
					free(stage[i]);
				stage_fill = 0;
				continue;
			}
		} else if (stage_fill == 10)
			errx(1, "%s: %zu: Too many lines for a stage (expected 10)", path, linenum);
		if (in_comment)
			free(line);
		else if (length < 10)
			errx(1, "%s: %zu: Line too short (expected 10 character, got %zd)", path, linenum, length);
		else if (length > 10)
			errx(1, "%s: %zu: Line too long (expected 10 character, got %zd)", path, linenum, length);
		else
			stage[stage_fill++] = line;
	}

	if (stage_fill) {
		if (stages_num == STAGES)
			errx(1, "%s: %zu: Too many stages (maximum %zu)", path, linenum, (size_t)STAGES);
		if (stage_fill < 10)
			errx(1, "%s: %zu: Not enough lines for a stage (expected 10, got %zu)", path, linenum, stage_fill);	
		char *stage_str;
		asprintf(&stage_str, "%s%s%s%s%s%s%s%s%s%s", stage[0], stage[1], stage[2], stage[3], stage[4], stage[5], stage[6], stage[7], stage[8], stage[9]);
		stages[stages_num++] = stage_str;
		for (size_t i = 0; i < stage_fill; i++)
			free(stage[i]);
	}

	if (!stages_num)
		errx(1, "%s: Bundle did not contain any stages", path);

	fclose(f);
}

int main(int argc, char *argv[]) {
	if (argc == 1) {
		stages_num = sizeof(builtin_stages) / sizeof(*builtin_stages);
		for (size_t i = 0; i < stages_num; i++) {
			stages[i] = builtin_stages[i];
		}
	} else if (argc == 2)
		load_stages(argv[1]);
	else {
		printf("Usage: %s [/path/to/stage-bundle]\n", argv[0]);
		return 1;
	}

	struct display_connection *connection = display_connect_default();
	if (!connection && errno == ECONNREFUSED)
		display_spawn(argc, argv);
	if (!connection)
		err(1, "Could not connect to display server");

	display_create_window(connection, main_window);
	display_resize_window(connection, main_window, window_width, window_height);
	display_show_window(connection, main_window);

	struct display_event_handlers handlers = {0};
	handlers.disconnect_handler = on_disconnect;
	handlers.keyboard_handler = on_keyboard;
	handlers.quit_handler = on_quit;
	handlers.resize_handler = on_resize;

	uint32_t palette_colours[] = {
		[BG] = make_color(128, 128, 255),
		[TEXT] = make_color(255, 255, 255),
		[PLAYER_OUTLINE] = make_color(0, 0, 0),
		[PLAYER_SKIN] = make_color(180, 100, 64),
		[PLAYER_EYE] = make_color(255, 255, 255),
		[GRASS_TOP] = make_color(0, 220, 0),
		[GRASS_BROWN] = make_color(200, 128, 0),
		[GRASS_DARK] = make_color(100, 64, 0),
		[FLAG_STEM] = make_color(0, 0, 0),
		[FLAG_FLAG] = make_color(255, 0, 0),
		[SWITCH_SWITCH] = make_color(200, 64, 0),
		[TILE1] = make_color(0, 255, 0),
		[TILE2] = make_color(0, 0, 255),
		[TILE3] = make_color(255, 0, 0),
		[TILE4] = make_color(255, 255, 0),
		[STAR_STAR] = make_color(255, 255, 0),
	};

	long fps_cap = 100;
	struct timespec min_dt = timespec_make(0, 1000L * 1000L * 1000L / fps_cap);
	struct timespec last_frame;
	clock_gettime(CLOCK_MONOTONIC, &last_frame);

	initialize();

	int fps = 0;
	while (game_running) {
		struct timespec now;
		clock_gettime(CLOCK_MONOTONIC, &now);
		struct timespec dt = timespec_sub(now, last_frame);
		if (timespec_lt(dt, min_dt)) {
			struct timespec sleep = timespec_sub(min_dt, dt);
			nanosleep(&sleep, NULL);
			continue;
		}

		if (last_frame.tv_sec != now.tv_sec) {
			char *title = NULL;
			asprintf(&title, "Switcher (%i fps)", fps);
			display_title_window(connection, main_window, title);
			free(title);
			fps = 0;
		} else
			fps++;

		update(now, dt);
		draw(now);

		size_t pixels;
		if (__builtin_mul_overflow(window_width, window_height, &pixels))
			errx(1, "Window size too big");
		uint32_t *framebuffer = calloc(pixels, sizeof(uint32_t));

		uint32_t shorter_side = window_height < window_width ? window_height : window_width;
		size_t y_offset = (window_height - shorter_side) / 2;
		size_t x_offset = (window_width - shorter_side) / 2;
		for (size_t y = 0; y < window_height; y++) {
			for (size_t x = 0; x < window_width; x++) {
				if (y < y_offset || x < x_offset ||
				    y >= y_offset + shorter_side || x >= x_offset + shorter_side)
					framebuffer[y * window_width + x] = make_color_a(0, 0, 0, 200);
				else {
					size_t pf_x = (x - x_offset) * PLAYFIELD_SIDE / shorter_side;
					size_t pf_y = (y - y_offset) * PLAYFIELD_SIDE / shorter_side;
					uint32_t colour = palette_colours[playfield[pf_y * PLAYFIELD_SIDE + pf_x]];
					framebuffer[y * window_width + x] = colour;
				}
			}
		}
		display_render_window(connection, main_window, 0, 0, window_width, window_height, framebuffer);
		free(framebuffer);

		while (!display_poll_event(connection, &handlers));
		
		last_frame = now;
	}

	display_disconnect(connection);

	return 0;
}
